/*
 Apache httpd module that uses RDMA to deliver data.

 This code is made available via the Apache License v2.0 (see LICENSE)
 and is Copyright 2007
		Dennis Dalessandro (dennis@osc.edu) - Ohio Supercomputer Center
		Pete Wyckoff (pw@osc.edu) - Ohio Supercomputer Center


 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <pthread.h>
#include <stdarg.h>
#include <time.h>


#include "clustercore/ccil_api.h"

#define CONTROL_BUF_SIZE 1024
#define CONN_BACK_LOG 10
#define MAX_POLL 10000000
#define SIGNAL_SIZE 10

#if 0
#define IW_VERBOSE_LEVEL 1
#define iwprintf(lvl,fmt,args...) \
    do { \
		if (lvl <= IW_VERBOSE_LEVEL) \
	    	iw_print_info(fmt,##args); \
    } while (0)
#else
#define iwprintf(lvl,fmt,args...) do { } while (0)
#endif


typedef enum {
	IDLE = 1,
	CONNECTED,
	OPEN,
	CLOSED,
} state_t;

typedef struct {
	//~ /*Important CCIL related members*/
	state_t			rnic_open;
	cc_rnic_handle_t	rnic_hndl; /*handle to the rnic device*/
	cc_cq_handle_t		cq_hndl; /*handle to the only completion queue we use*/
	cc_uint32_t		cq_depth; /*depth of the completion queue*/
	cc_pdid_t			prot_domain_id; /*id fo the protection domain*/
	cc_qp_handle_t		qp_hndl; /*qp handle*/
	cc_netaddr_t 		host_type_addr;
	unsigned short	port;
	cc_qp_id_t		qp_id; /*id of the qp*/
	cc_ep_handle_t		listen_ep_hndl;  /*end point handle for the listen call */
	cc_ep_handle_t		cr_ep_hndl;  /*end point handle for the accept call*/
	//~ cc_eh_ce_handler_id_t	cq_handler_id; /*used to id the function we use as the completion queue handler*/
	cc_rq_wr_t 		rq_wr;		/*recv work request record*/
	cc_sq_wr_t 		sq_wr;		/*send work request record*/
	cc_sq_wr_t 		rdma_sq_wr;	/* send work request record*/
	cc_data_addr_t 	recv_sgl;	/*recv single SGE*/
	cc_data_addr_t 	send_sgl;	/*send single SGE*/
	cc_data_addr_t 	rdma_sgl;	/*send single SGE*/
	cc_stag_index_t	recv_stag_index; /*STag for recv buffer*/
	cc_stag_index_t	send_stag_index; /*STag for send buffer*/
	cc_stag_index_t	rdma_stag_index; /*STag for RDMA buffer*/
	//~ cc_stag_index_t		last_msg_stag_index; /*stag for last message buffer*/
	cc_stag_t			remote_stag;	/*remote guys RDMA buffer STAG*/
	cc_uint64_t		remote_to;	/*remote guys RDMA buffer TO*/
	//~ cc_uint32_t		remote_len;	// remote guys LEN
	unsigned char	*recv_buf; /*receive buffer, needs to be Malloc'd*/
	unsigned char	*send_buf; /*send buffer, needs to be Malloc'd*/
	unsigned char	*rdma_buf; /*send registered buffer*/
	int				mapped; /*used to tell if rdma_buf is memory mapped or not*/
	int				rdma_bufsize;
	cc_uint64_t		rdma_to;
	char 			ip_address[20];
	//~ unsigned char			*last_msg;  /*buffer to hold last message of each window for bw test*/

	//~ /*Application related members*/
	//~ int			am_server; /*whether server or client*/
	state_t			state;	   /*used for cond/signalling*/
	pthread_mutex_t 	LOCK;
	pthread_cond_t 	COND;
	pthread_mutex_t	READY;
	pthread_mutex_t	mutex_done;
	pthread_mutex_t	EVENT;
	int 				outstanding_cqe;
	int				static_buffer;
	int 				not_tcp;  /*set this if its not TCP so we don't try to deallocate nothing*/
	int				recv_allocated;
	int				send_allocated;
	int				rdma_allocated;
} info;




void rnic_async_event_handler(cc_rnic_handle_t rnic_hndl, cc_event_record_t* event_record, void *user_context);
void am_error(const char *msg, cc_status_t code);
int open_rdma_dev(cc_rnic_handle_t *rnic_hndl, int flag, void *user_context);
int rdma_init(info *rdma_info);
int rdma_allocate_resources(info *rdma_info, int rdma_buf_size);
int rdma_connect(info *rdma_info, char *server, short unsigned int port);
int rdma_listen_connection(info *rdma_info);
int rdma_wait_connection(info *rdma_info);
int rdma_deallocate_resources(info *rdma_info);
int rdma_send_file(info *rdma_info, char *filename, int size); /*needs to take buffer or fd 2 funcs*/
void cqe_evh(cc_rnic_handle_t rnic, cc_cq_handle_t cqh, void *user_context);
int rdma_poll_cq(info *rdma_info, cc_wc_t *wc, u_int32_t polls, u_int32_t sleep_time);
void iw_print_info(const char *fmt, ...);
int send_rdma_buffer_info(int socket_fd, int rdma_buf_size, cc_stag_index_t stag, cc_uint64_t to);
int rdma_map_file(info *rdma_info, char *filename, int size);
int rdma_read_file(info *rdma_info, int size);
int rdma_send_signal(info *rdma_info);
int rdma_wait_signal(info *rdma_info);
int rdma_write_from_file(info *rdma_info, int fd, int offset, int len);
int rdma_write_from_buffer(info *rdma_info, void *buffer, int len);
int rdma_reap_completions(info *rdma_info);
int rdma_read_map_file(info *rdma_info, int fd, int offset, int size);
int rdma_read_map_buffer(info *rdma_info, void *buffer, int size);
int rdma_write_from_file_static(info *rdma_info, int fd, int offset, int len);
int rdma_read_map_file_static(info *rdma_info, int fd, int offset, int size);
int rdma_cleanup(info *rdma_info);
int rdma_zero_info(info *rdma_info);
