/*
 Apache httpd module that uses RDMA to deliver data.

 This code is made available via the Apache License v2.0 (see LICENSE)
 and is Copyright 2007
		Dennis Dalessandro (dennis@osc.edu) - Ohio Supercomputer Center
		Pete Wyckoff (pw@osc.edu) - Ohio Supercomputer Center


 */

#include "openfab.h"

 #define IGNORE __attribute__ ((unused))

int OK_TO_QUIT;

int PAGE_THRESHOLD = MAX_PAGE;
int MAX_OPS = MAX_OUTSTANDING;


/*---------------------------*/
/*Apache and Shared Functions*/
/*---------------------------*/

#ifdef OF_SOFTWARE
struct ibv_comp_channel *ibv_create_comp_channel(struct ibv_context *context){
	return NULL;
}

int ibv_destroy_comp_channel(struct ibv_comp_channel *channel){
	return 0;
}

int ibv_get_cq_event(struct ibv_comp_channel *channel,
		     struct ibv_cq **cq, void **cq_context){
	return 0;
}

void ibv_ack_cq_events(struct ibv_cq *cq, unsigned int nevents){
	return;
}

#endif

/*NEEDED FOR TESTING PROGRAMS AND INTERNAL FUNCTIONS*/
#define CASE(e)  case e: s = #e; break

const char *my_openib_wc_status_string(int status)
{
/*taken from ardma by Pete Wyckoff*/
	const char *s = "(UNKNOWN)";

	switch (status) {
		CASE(IBV_WC_SUCCESS);
		CASE(IBV_WC_LOC_LEN_ERR);
		CASE(IBV_WC_LOC_QP_OP_ERR);
		CASE(IBV_WC_LOC_EEC_OP_ERR);
		CASE(IBV_WC_LOC_PROT_ERR);
		CASE(IBV_WC_WR_FLUSH_ERR);
		CASE(IBV_WC_MW_BIND_ERR);
		CASE(IBV_WC_BAD_RESP_ERR);
		CASE(IBV_WC_LOC_ACCESS_ERR);
		CASE(IBV_WC_REM_INV_REQ_ERR);
		CASE(IBV_WC_REM_ACCESS_ERR);
		CASE(IBV_WC_REM_OP_ERR);
		CASE(IBV_WC_RETRY_EXC_ERR);
		CASE(IBV_WC_RNR_RETRY_EXC_ERR);
		CASE(IBV_WC_LOC_RDD_VIOL_ERR);
		CASE(IBV_WC_REM_INV_RD_REQ_ERR);
		CASE(IBV_WC_REM_ABORT_ERR);
		CASE(IBV_WC_INV_EECN_ERR);
		CASE(IBV_WC_INV_EEC_STATE_ERR);
		CASE(IBV_WC_FATAL_ERR);
		CASE(IBV_WC_GENERAL_ERR);
	}
	return s;
}

void openfab_error(const char *msg, int code)
{
	/*print error message and string form error code passed in*/
	char error_string[256];
	int length_of_error_string = 256;

	strerror_r(code, error_string, length_of_error_string);
   	fprintf(stderr,"----------ERROR----------\n");
	fprintf(stderr, "%s : %s\n", msg, error_string);
	fprintf(stderr,"----------ERROR----------\n");
}

void iw_print_info(const char *fmt, ...)
{
/*Debugging function prints message in nice format*/
    va_list ap;
    fprintf(stderr, "..........[iwarp_module]:");
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, ".\n");
    fflush(stderr);
}



int rdma_init(info *rdma_info)
{
/*Init the RNIC*/
	struct ibv_qp_init_attr qp_init_attr;
	int cqe_num;
	struct ibv_device_attr device_attr;
	int ret;
	struct hostent *host;
	struct sockaddr_in sock_addr;
	struct rdma_cm_event *event;

	/*resolve either the IP of the other guy or our own
	just to get the ibv_hndl*/
	iwprintf(5, "%s() Entry", __func__);
	iwprintf(5, "%s() Looking up %s", __func__, rdma_info->ip_address);

	if(rdma_info->port_override != 0)
		rdma_info->port = rdma_info->port_override;
        else{
		srand((unsigned)time(NULL));
		rdma_info->port = rand() % 9999;
		if(rdma_info->port < 1000)
			rdma_info->port += 1000;
	}

	iwprintf(9, "%s() Using RDMA Port %d", __func__, rdma_info->port);

	rdma_info->cma_id = NULL;
	rdma_info->cma_listen_id = NULL;
	rdma_info->called_close = 0;

	memset(&sock_addr, 0, sizeof(sock_addr));
	sock_addr.sin_family = AF_INET;
	sock_addr.sin_port = htons(rdma_info->port);

	rdma_info->rdma_evt_channel = rdma_create_event_channel();

	host = gethostbyname(rdma_info->ip_address);
	if (!host){
		iwprintf(0, "Unable to gethostbyname");
		openfab_error("Unable to resolve host", errno);
		return -1;
	}

	memcpy(&sock_addr.sin_addr, host->h_addr_list[0], host->h_length);

	ret = rdma_create_id(rdma_info->rdma_evt_channel, &(rdma_info->cma_id),
					NULL, RDMA_PS_TCP);
	if (ret){
		iwprintf(0, "Can't create RDMA ID");
		openfab_error("Unable to create RDMA ID", -ret);
		return -1;
	}

	ret = rdma_resolve_addr(rdma_info->cma_id, NULL,
				(struct sockaddr *) &sock_addr,
				3000);  /* 3 sec timeout */
	if (ret){
		iwprintf(0, "Can't resolve address");
		openfab_error("Unable to resolve address", -ret);
		return -1;
	}

	ret = rdma_get_cm_event(rdma_info->rdma_evt_channel, &event);
	if (ret) {
		if (ret == -1){
			iwprintf(0, "can't get CM event got -1");
			openfab_error("Unable to get a CM event", errno);
			return -1;
		}
		else{
			iwprintf(0, "can't get CM event got something other than -1");
			openfab_error("Unable to get a CM event", -ret);
			return -1;
		}
	}

	if(event->event != RDMA_CM_EVENT_ADDR_RESOLVED){
		iwprintf(0,"inavlid event received, expecting  %d RDMA_CM_EVENT_ADDR_RESOLVED, recevd %d ",
					RDMA_CM_EVENT_ADDR_RESOLVED, event->event);
		if (event->event == RDMA_CM_EVENT_ADDR_ERROR)
			iwprintf(0, "The invalid event received is RDMA_CM_EVENT_ADDR_ERROR");
		openfab_error("Invalid event received", errno);
		ret = rdma_ack_cm_event(event);
		if (ret)
			openfab_error("Unable to release event", -ret);
		return -1;
	}

	rdma_info->ibv_hndl = event->id->verbs;

	ret = rdma_ack_cm_event(event);
	if(ret){
		openfab_error("Unable to release event", -ret);
		return -1;
	}

	rdma_info->rnic_state = IDLE;

	/*allocate protection domain*/
	rdma_info->prot_id = ibv_alloc_pd(rdma_info->ibv_hndl);
	if (!rdma_info->prot_id){
		openfab_error("Unable to create PD", errno);
		return -1;
	}

	/*create completion queues*/
	ret = ibv_query_device(rdma_info->ibv_hndl, &device_attr);
	if (ret < 0){
		openfab_error("Unable to query device to figure out max CQs", errno);
		return -1;
	}

	iwprintf(5, "%s()Max %d completion queue entries", __func__,
				device_attr.max_cqe);

	/*
	can't allocate a full max_cqe for some reason when using Chelsio HW
	use something much less
	*/
	cqe_num = 100;  /*XXX TODO: come up with a better value*/

	rdma_info->cq_channel = NULL;

	rdma_info->cq_channel = ibv_create_comp_channel(rdma_info->ibv_hndl);

	rdma_info->cq_hndl = ibv_create_cq(rdma_info->ibv_hndl, cqe_num, NULL,
				rdma_info->cq_channel, 0);
	if (!rdma_info->cq_hndl){
		openfab_error("Unable to create CQ", errno);
		return -1;
	}

	/*now to create the QP*/
	memset(&qp_init_attr, 0, sizeof(qp_init_attr));
	qp_init_attr.send_cq =  rdma_info->cq_hndl;
	qp_init_attr.recv_cq  = rdma_info->cq_hndl;
	qp_init_attr.cap.max_send_wr = 100;  /* outstanding WQEs */
	qp_init_attr.cap.max_recv_wr = 100; /*XXX TODO: Come up with better value*/
	qp_init_attr.cap.max_send_sge = 1;  /* scatter/gather entries */
	qp_init_attr.cap.max_recv_sge = 1;
	qp_init_attr.qp_type = IBV_QPT_RC;
	/* only generate completion queue entries if requested */
	qp_init_attr.sq_sig_all = 0;
	ret = rdma_create_qp(rdma_info->cma_id, rdma_info->prot_id, &qp_init_attr);
	if (ret){
		openfab_error("unable to create qp", -ret);
		iwprintf(0, "Unable to create QP");
		return -1;
	}
	rdma_info->qp_hndl = rdma_info->cma_id->qp;

	iwprintf(5, "%s() Exit", __func__);


    return 0;
}



int rdma_deallocate_resources(info *rdma_info)
{
/*deallocate/free memory and memory regions*/

	int ret;

	iwprintf(5, "%s() Entry", __func__);

	iwprintf(5, "%s() Deallocting RDMA members", __func__);
	if (rdma_info->rdma_reg) {
		iwprintf(4, "%s() DeReg MR", __func__);
		ret = ibv_dereg_mr(rdma_info->rdma_mr);
		if (ret){
			iwprintf(0, "Can't dereg rdma_mr");
			openfab_error("Unable to deregister rdma mem region", ret);
			return -1;
		}
		rdma_info->rdma_reg = 0;
	}

	if (rdma_info->rdma_mapped && rdma_info->rdma_malloced) {
		openfab_error("Unable to unmap and free same buffer", -1);
		return -1;
	}

	if (rdma_info->rdma_mapped){
		iwprintf(4, "%s() Unmap mem", __func__);
		ret = munmap(rdma_info->rdma_buf, rdma_info->rdma_bufsize);
		if (ret){
			iwprintf(0, "Unable to unmap buffer");
			openfab_error("Unable to unmap buffer", errno);
			return -1;
		}
		rdma_info->rdma_mapped = 0;
	}

	if (rdma_info->rdma_malloced) {
		iwprintf(4, "%s() Free mem", __func__);
		free(rdma_info->rdma_buf);
		rdma_info->rdma_malloced = 0;
	}

	iwprintf(5, "%s() Deallocting RECV members", __func__);

	if (rdma_info->recv_reg) {
		iwprintf(4, "%s() DeReg RECV MR", __func__);
		ret = ibv_dereg_mr(rdma_info->recv_mr);
		if (ret){
			iwprintf(0, "Unable to dereg recv mem");
			openfab_error("Unable to deregister recv mem region", ret);
			return -1;
		}
		rdma_info->recv_reg = 0;
	}

	if (rdma_info->recv_malloced) {
		iwprintf(2, "%s() Free recv mem", __func__);
		free(rdma_info->recv_buf);
		rdma_info->recv_malloced = 0;
	}

	iwprintf(5, "%s() Deallocting SEND members", __func__);

	if (rdma_info->send_reg) {
		iwprintf(4, "%s() DeReg Send MR", __func__);
		ret = ibv_dereg_mr(rdma_info->send_mr);
		if (ret){
			iwprintf(0, "Unable to deregister send mem region");
			openfab_error("Unable to deregister send mem region", ret);
			return -1;
		}
		rdma_info->send_reg = 0;
	}


	if (rdma_info->send_malloced) {
		iwprintf(2, "%s() Free send mem", __func__);
		free(rdma_info->send_buf);
		rdma_info->send_malloced = 0;
	}

	rdma_info->called_close = 1;

	iwprintf(5, "%s() Exit", __func__);
	return 0;

}


int rdma_do_connect( info *rdma_info,  IGNORE char *server, IGNORE short unsigned int port)
/*establish rdma connection back to client*/
{

	int ret;
	struct rdma_cm_event *event;
	struct rdma_conn_param conn_param = {
		.responder_resources = 1,
		.initiator_depth = 1,
		.retry_count = 5,
	};

	iwprintf(5, "%s() Entry", __func__);

	rdma_info->cma_listen_id = NULL;

	/*now we can resolve the route to the web client*/
	ret = rdma_resolve_route(rdma_info->cma_id, 30000);
	if (ret){
		openfab_error("Unable to resolve route", -ret);
		return -1;
	}

	/*get the route resolved event*/
	ret = rdma_get_cm_event(rdma_info->rdma_evt_channel, &event);
	if (ret) {
		if (ret == -1){
			openfab_error("Unable to get a CM event", errno);
			return -1;
		} else {
			openfab_error("Unable to get a CM event", -ret);
			return -1;
		}
	}

	if (event->event != RDMA_CM_EVENT_ROUTE_RESOLVED){
		iwprintf(0,"inavlid event espect %d got recevd %d",
				RDMA_CM_EVENT_ROUTE_RESOLVED, event->event);
		openfab_error("Invalid event received", errno);
		return -1;
	}

	iwprintf(5, "%s() Route is resolved", __func__);

	/*clear route resolved event*/
	ret = rdma_ack_cm_event(event);
	if (ret){
		openfab_error("Unable to release event", -ret);
		return -1;
	}

	/*attempt to make the connection*/
	ret = rdma_connect(rdma_info->cma_id, &conn_param);
	if (ret){
		openfab_error("Unable to connect", -ret);
		return -1;
	}

	iwprintf(5, "%s() Making the connection waiting for evt", __func__);

	/*now check for connection made event*/
	ret = rdma_get_cm_event(rdma_info->rdma_evt_channel, &event);
	if (ret) {
		if (ret == -1){
			openfab_error("Unable to get a CM event", errno);
			return -1;
		} else {
			openfab_error("Unable to get a CM event", -ret);
			return -1;
		}
	}

	if(event->event != RDMA_CM_EVENT_ESTABLISHED){
		iwprintf(0,"inavlid event expecting  %d recevd %d",
				RDMA_CM_EVENT_ESTABLISHED, event->event);
		openfab_error("Invalid event received", errno);
		return -1;
	}

	/*clear connection made event*/
	ret = rdma_ack_cm_event(event);
	if(ret){
		openfab_error("Unable to release event", -ret);
		return -1;
	}

	rdma_info->rnic_state = CONNECTED;

	iwprintf(5, "%s() Exit", __func__);

	return 0;
}


int rdma_zero_info(info *rdma_info)
/*init some important initial values to 0*/
{

	/*anything else that needs to be initialized for EACH connection do it in here*/
	rdma_info->not_tcp = 0;
	rdma_info->rdma_mapped = 0;
	rdma_info->rdma_malloced = 0;
	rdma_info->rdma_reg = 0;

	return 0;
}


int rdma_write_from_file(info *rdma_info, int fd,  int offset,  int len)
/*do an RDMA write from an open file - non php requests*/
{
	int ret;



	iwprintf(5, "%s() Entry", __func__);

	/*since we are using mmpa offset should be a multiple of page size if not warn, and continue anyway*/
	if(offset / getpagesize() != 0){
		fprintf(stderr, "WARNING OFFSET IS NOT A MULTIPLE OF PAGE SIZE\n");
	}


	#ifdef SENDFILE

	iwprintf(5, "%s() Using Prot Domain %p", __func__, rdma_info->prot_id);


	iwprintf(5, "%s() From QP qp_hndl->pd = %p", __func__,rdma_info->qp_hndl->pd);

	ret = u_rsf_init(&rdma_info->rsf_descriptor, rdma_info->qp_hndl,
				rdma_info->cq_hndl, rdma_info->cq_channel,
				rdma_info->remote_stag, rdma_info->remote_to);
	if(ret != 0){
		openfab_error("Unable to init rsf lib", ret);

		return -1;
	}
	ret = u_rsf_modify_threshold(&rdma_info->rsf_descriptor, 256);
	if(ret != 0){
		openfab_error("Unable to modify threshold", ret);
		return -1;
	}

	ret = u_rsf_modify_outstanding(&rdma_info->rsf_descriptor, 6);
	if(ret != 0){
		openfab_error("Unable to modify outsanding ops", ret);
		return -1;
	}

	iwprintf(5,"%s() Page Threshold = %d", __func__,
			u_rsf_query_threshold(&rdma_info->rsf_descriptor));
	iwprintf(5, "%s() Max Outstanding Ops = %d", __func__,
			u_rsf_query_outstanding(&rdma_info->rsf_descriptor));


	ret = rdma_sendfile(fd, &rdma_info->rsf_descriptor, offset, len);
	if (ret != (ssize_t) len) {
		perror("Unable to complete SENDFILE operation\n");
		return -1;
	}

	iwprintf(5, "%s() Sendfile complete now for fini", __func__);

	ret = u_rsf_fini(&rdma_info->rsf_descriptor);
	if (ret !=0 ){
		openfab_error("Unable to fini", ret);
		return -1;
	}


	#else
	struct ibv_send_wr *bad_wr;
	/*For mmap*/
	rdma_info->rdma_buf = mmap(0, len, PROT_READ, MAP_PRIVATE, fd, offset);
	if(rdma_info->rdma_buf == (void *) -1){ /*mem map failed*/
		perror("Map of file failed with");
		return -1;
	}

	rdma_info->rdma_malloced = 0;
	rdma_info->rdma_mapped = 1;

	/*register rdma buffer*/
	rdma_info->rdma_mr = ibv_reg_mr(rdma_info->prot_id, rdma_info->rdma_buf, len,  0);
	if (!rdma_info->rdma_mr){
		openfab_error("Unable to register rdma buffer", errno);
		return -1;
	}
	rdma_info->rdma_reg = 1;

	iwprintf(5, "%s() RDMA Buffer", __func__);
	iwprintf(5, "%s() lkey %d", __func__, rdma_info->rdma_mr->lkey);
	iwprintf(5, "%s() rkey %d", __func__, rdma_info->rdma_mr->rkey);

	rdma_info->rdma_to = uint64_from_ptr(rdma_info->rdma_buf);
	rdma_info->rdma_stag_index = rdma_info->rdma_mr->rkey;

	iwprintf(10,"%s() buffer is pinned and ready to go", __func__);

	/*build up the server work request*/
	rdma_info->rdma_sge.addr = uint64_from_ptr(rdma_info->rdma_buf);
	rdma_info->rdma_sge.length = len;
	rdma_info->rdma_sge.lkey = rdma_info->rdma_mr->lkey;

	memset(&(rdma_info->rdma_wr), 0, sizeof(rdma_info->rdma_wr));

	rdma_info->rdma_wr.wr_id = uint64_from_ptr(&(rdma_info->rdma_wr));
	rdma_info->rdma_wr.sg_list = &(rdma_info->rdma_sge);
	rdma_info->rdma_wr.num_sge = 1;
	rdma_info->rdma_wr.wr.rdma.remote_addr = rdma_info->remote_to;
	rdma_info->rdma_wr.wr.rdma.rkey = rdma_info->remote_stag;
	rdma_info->rdma_wr.opcode = IBV_WR_RDMA_WRITE;
	rdma_info->rdma_wr.send_flags = IBV_SEND_SIGNALED;

	iwprintf(5,"%s() going to write to STag = %d and TO = %lld",
				__func__, rdma_info->remote_stag,
				rdma_info->remote_to);

	iwprintf(5, "%s() Requesting notification of next event on CQ",
								__func__);
	ret = ibv_req_notify_cq(rdma_info->cq_hndl, 0);
	if (ret){
		openfab_error("Unable to get notication for next cq", ret);
		return -1;
	}
	iwprintf(5, "%s() Posting the send", __func__);
	ret = ibv_post_send(rdma_info->qp_hndl, &(rdma_info->rdma_wr), &bad_wr);
	if (ret){
		openfab_error("Unable to post to SQ", ret);
		return -1;
	}

	rdma_info->rdma_bufsize = len;
	rdma_info->cq_count = 1;

	#endif

	iwprintf(5, "%s() Exit", __func__);

	return 0;

}


int setup_read_buffer_file(info *rdma_info, int fd, int offset, int len)
/*prep buffer so client can do an rdma read and post recv to wait on*/
{
	int ret;
	struct ibv_recv_wr *bad_wr;

	iwprintf(5, "%s() Entry", __func__);

	/*since we are using mmpa offset should be a multiple of page size if not warn, and continue anyway*/
	if(offset / getpagesize() != 0){
		fprintf(stderr, "WARNING OFFSET IS NOT A MULTIPLE OF PAGE SIZE\n");
	}

	/*For mmap*/
	rdma_info->rdma_buf = mmap(0, len, PROT_READ, MAP_PRIVATE, fd, offset);
	if(rdma_info->rdma_buf == (void *) -1){ /*mem map failed*/
		perror("Map of file failed with");
		return -1;
	}

	rdma_info->rdma_malloced = 0;
	rdma_info->rdma_mapped = 1;

	/*register rdma buffer*/
	rdma_info->rdma_mr = ibv_reg_mr(rdma_info->prot_id, rdma_info->rdma_buf, len,  0);
	if (!rdma_info->rdma_mr){
		openfab_error("Unable to register rdma buffer", errno);
		return -1;
	}
	rdma_info->rdma_reg = 1;

	rdma_info->rdma_to = uint64_from_ptr(rdma_info->rdma_buf);
	rdma_info->rdma_stag_index = rdma_info->rdma_mr->rkey;
	rdma_info->rdma_bufsize = len;

	iwprintf(5, "%s() RDMA Buffer", __func__);
	iwprintf(5, "%s() STag %d", __func__, rdma_info->rdma_stag_index);
	iwprintf(5, "%s() Length %d", __func__, rdma_info->rdma_bufsize);
	#ifdef __x86_64__
	iwprintf(5, "%s() TO %ld", __func__, rdma_info->rdma_to);
	#else
	iwprintf(5, "%s() TO %lld", __func__, rdma_info->rdma_to);
	#endif

	iwprintf(10,"%s() buffer is pinned and ready to go", __func__);


	/*now need to register a receive buffer to wait for response from
	client*/
	rdma_info->recv_buf = malloc(SIGNAL_SIZE);
	if (!rdma_info->recv_buf) {
		openfab_error("Unable to malloc memory for ACK buf", errno);
		return -1;
	}

	rdma_info->recv_mr = ibv_reg_mr(rdma_info->prot_id, rdma_info->recv_buf,
				SIGNAL_SIZE,  0);
	if (!rdma_info->recv_mr){
		openfab_error("Unable to register recv buffer", errno);
		return -1;
	}
	rdma_info->recv_reg = 1;

	/*buld up the recv work request*/
	rdma_info->recv_sge.addr = uint64_from_ptr(rdma_info->recv_buf);
	rdma_info->recv_sge.length = SIGNAL_SIZE;
	rdma_info->recv_sge.lkey = rdma_info->recv_mr->lkey;

	memset(&(rdma_info->recv_wr), 0, sizeof(rdma_info->recv_wr));
	rdma_info->recv_wr.wr_id = uint64_from_ptr(&rdma_info->recv_wr);
	rdma_info->recv_wr.sg_list = &(rdma_info->recv_sge);
	rdma_info->recv_wr.num_sge = 1;
	rdma_info->recv_wr.next = NULL;


	iwprintf(5, "%s() Requesting notification of next event on CQ",
								__func__);
	ret = ibv_req_notify_cq(rdma_info->cq_hndl, 0);
	if (ret){
		openfab_error("Unable to get notication for next cq", ret);
		return -1;
	}

	ret = ibv_post_recv(rdma_info->qp_hndl, &(rdma_info->recv_wr), &bad_wr);
	if (ret){
		openfab_error("Can't post RECV", errno);
		return -1;
	}

	iwprintf(5, "%s() Recv is posted", __func__);


	rdma_info->cq_count = 1;

	iwprintf(5, "%s() Exit", __func__);

	return 0;

}
int rdma_write_from_buffer( info *rdma_info,  void *buffer,  int len)
/*do an RDMA write from a buffer - php requests*/
{

	int ret;
	struct ibv_send_wr *bad_wr;

	iwprintf(5, "Entry", __func__);

	iwprintf(5, "%s() Going to write from %p a total len of %d bytes",
					__func__, buffer, len);

	/*register rdma buffer*/
	rdma_info->rdma_mr = ibv_reg_mr(rdma_info->prot_id, buffer, len,
			IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
	if (!rdma_info->rdma_mr){
		openfab_error("Unable to register rdma buffer", errno);
		return -1;
	}
	rdma_info->rdma_reg = 1;


	iwprintf(5, "%s() RDMA Buffer", __func__);
	iwprintf(5, "%s() lkey %d", __func__, rdma_info->rdma_mr->lkey);
	iwprintf(5, "%s() rkey %d", __func__, rdma_info->rdma_mr->rkey);

	rdma_info->rdma_to = uint64_from_ptr(rdma_info->rdma_buf);
	rdma_info->rdma_stag_index = rdma_info->rdma_mr->rkey;

	/*build up the server work request*/
	rdma_info->rdma_sge.addr = uint64_from_ptr(buffer);
	rdma_info->rdma_sge.length = len;
	rdma_info->rdma_sge.lkey = rdma_info->rdma_mr->lkey;

	memset(&(rdma_info->rdma_wr), 0, sizeof(rdma_info->rdma_wr));

	rdma_info->rdma_wr.wr_id = uint64_from_ptr(&(rdma_info->rdma_wr));
	rdma_info->rdma_wr.sg_list = &(rdma_info->rdma_sge);
	rdma_info->rdma_wr.num_sge = 1;
	rdma_info->rdma_wr.wr.rdma.remote_addr = rdma_info->remote_to;
	rdma_info->rdma_wr.wr.rdma.rkey = rdma_info->remote_stag;
	rdma_info->rdma_wr.opcode = IBV_WR_RDMA_WRITE;
	rdma_info->rdma_wr.send_flags = IBV_SEND_SIGNALED;

	#ifdef __x86_64__
	iwprintf(5,"%s() Going to write to STag = %d and TO = %ld",
				__func__,
				rdma_info->remote_stag,
				rdma_info->remote_to);
	#else
	iwprintf(5,"%s() Going to write to STag = %d and TO = %lld",
				__func__,
				rdma_info->remote_stag,
				rdma_info->remote_to);
	#endif

	iwprintf(5, "%s() Requesting notification next event on CQ", __func__);
	ret = ibv_req_notify_cq(rdma_info->cq_hndl, 0);
	if (ret){
		openfab_error("Unable to get notication for next cq", ret);
		return -1;
	}

	iwprintf(10,"%s() posting send", __func__);

	ret = ibv_post_send(rdma_info->qp_hndl, &(rdma_info->rdma_wr), &bad_wr);
	if (ret){
		openfab_error("Unable to post to SQ", ret);
		return -1;
	}

	rdma_info->cq_count = 1;

	return 0;
}


int rdma_reap_completions(info *rdma_info, uint32_t num_events)
/*using blocking mode poll for desired num completions*/
{

	int ret = 0;
	struct ibv_wc wc;
	struct ibv_cq *cq = NULL;
	void *cq_context;
	uint32_t i;

	iwprintf(5, "%s() Entry", __func__);

	iwprintf(5, "%s() Getting %d events", __func__, num_events);

	for (i=0; i<num_events; i++){

		ret = ibv_poll_cq(rdma_info->cq_hndl, 1, &wc);
		if (ret < 0){
			openfab_error("Error getting completion event", errno);
			return ret;
		} else if(ret == 0){ /*nothign there so we need to block*/
			iwprintf(5, "%s() Going to block until someting arrives",
						__func__);
			/*block waiting for completion*/
			ret = ibv_get_cq_event(rdma_info->cq_channel, &cq, &cq_context);

			/*now poll, somethign SHOULD BE THERE or else BIG Trouble*/
			iwprintf(5, "%s() Done blocking", __func__);
			ret = ibv_poll_cq(cq, 1, &wc);
			if(ret <= 0){
				openfab_error("Error getting cq event", errno);
				return -1;
			}
			/*have to ACK the event still*/
			ibv_ack_cq_events(cq,1);
		}
		else{
			iwprintf(5, "%s() Got a completion event on the first try",
						__func__);
		}
		/*got a completion event lets check it and be done with it*/
		iwprintf(5, "%s() Check the completion event", __func__);
		if(wc.status != IBV_WC_SUCCESS){
			iwprintf(0, "%s() Invaild WC status received %s", __func__,
						my_openib_wc_status_string(wc.status));
			openfab_error("Invalid WC status", 0);
			return -1;
		}
		iwprintf(5, "%s() Got CQ event", __func__);
		if(wc.opcode == IBV_WC_SEND)
			iwprintf(10, "%s() Op is SEND", __func__);
		else if(wc.opcode == IBV_WC_RECV)
			iwprintf(10, "%s() Op is RECV", __func__);
		else if(wc.opcode == IBV_WC_RDMA_WRITE)
			iwprintf(10, "%s() Op is RDMA Write", __func__);
		else{
			openfab_error("invalid opcode", -1);
			return -1;
		}

		iwprintf(5, "%s() Requesting notification of next event on CQ",
								__func__);
		ret = ibv_req_notify_cq(rdma_info->cq_hndl, 0);
		if (ret){
			openfab_error("Unable to get notication for next cq", ret);
			return -1;
		}


	}
	ret = 0;
	iwprintf(5, "%s() Exit", __func__);
	return ret;
}










int rdma_cleanup( info *rdma_info)
/*tear down pd,cq,qp,cma_ids,event channels etc and kill connection*/
{

	int ret;
	int err_val = 0;
	struct rdma_cm_event *event;

	iwprintf(5, "%s() Entry", __func__);

	iwprintf(2, "%s() Getting disconnect event", __func__);
	rdma_get_cm_event(rdma_info->rdma_evt_channel, &event);
	if (event->event != RDMA_CM_EVENT_DISCONNECTED) {
		iwprintf(0, "Unable to get disconnected event");
		openfab_error("Unable to get disconnected event", errno);
		err_val = -1;
	} else {
		iwprintf(2, "%s() Got disconnect event", __func__);
	}

	ret = rdma_ack_cm_event(event);
	if(ret) {
		openfab_error("Unable to release event", -ret);
		return -1;
	}

	iwprintf(5, "%s() Got disconnect event and ackd it", __func__);

	iwprintf(5, "%s() Destroying QP", __func__);
	rdma_destroy_qp(rdma_info->cma_id);

	iwprintf(5, "%s() Destroying CMA ID", __func__);
	ret = rdma_destroy_id(rdma_info->cma_id);
	if (ret){
		openfab_error("Unable to destroy CMA ID", -ret);
		err_val = -1;
	}

	if (rdma_info->cma_listen_id != NULL){
		iwprintf(5, "%s() Destroying CMA Listen ID", __func__);
		ret = rdma_destroy_id(rdma_info->cma_listen_id);
		if(ret){
			openfab_error("Unable to destroy CMA listen ID", -ret);
			err_val = -1;
		}
	}

	iwprintf(5, "%s() Deallocating PD", __func__);
	ret = ibv_dealloc_pd(rdma_info->prot_id);
	if (ret){
		openfab_error("Unable to deallocate Protection domain", ret);
		fprintf(stderr, "PD dealloc failed for port %d\n", rdma_info->port);
		err_val = -1;
	}

	iwprintf(5, "%s() Destroying Completion Queue", __func__);
	ret = ibv_destroy_cq(rdma_info->cq_hndl);
	if (ret){
		openfab_error("Unable to destroy Completion queue", ret);
		err_val = -1;
	}

	iwprintf(5, "%s() Attempting to destroy comp channel", __func__);
	ret =  ibv_destroy_comp_channel(rdma_info->cq_channel);
	if(ret){
		openfab_error("Unable to destroy comp channel", -ret);
		err_val = -1;
	}

	/*might as well do this as well*/
	iwprintf(5, "%s() Destroy RDMA EVENT channel", __func__);
	rdma_destroy_event_channel(rdma_info->rdma_evt_channel);

	iwprintf(5, "%s() cleaning up RDMA resources", __func__);

	iwprintf(5, "%s() Exit", __func__);

	return err_val;



}



/*----------------------------*/
/*Additional Client Only Funcs*/
/*----------------------------*/
int rdma_listen_connection( info *rdma_info)
/*Client function to enable listening for an rdma connection from server*/
{
	unsigned short port;
	struct sockaddr_in sock_addr;
	int ret;

	iwprintf(5, "%s() Entry", __func__);

	port = rdma_info->port;
	iwprintf(5, "%s() Using Port #%d", __func__, port);

	memset(&sock_addr, 0, sizeof(sock_addr));
	sock_addr.sin_family = AF_INET;;
	sock_addr.sin_port = htons(port);


	rdma_info->rdma_evt_channel = rdma_create_event_channel();

	sock_addr.sin_addr.s_addr = INADDR_ANY;
	#ifdef OLDVERS
	ret = rdma_create_id(rdma_info->rdma_evt_channel,
				&(rdma_info->cma_listen_id),NULL);
	#else
	ret = rdma_create_id(rdma_info->rdma_evt_channel,
				&(rdma_info->cma_listen_id), NULL, RDMA_PS_TCP);
	#endif
	if (ret){
		openfab_error("Unable to create rdma id", -ret);
		return -1;
	}

	ret = rdma_bind_addr(rdma_info->cma_listen_id,
				(struct sockaddr *) &sock_addr);
	if (ret) {
		if (ret == -1)
			openfab_error("can't bind address", errno);
		else
			openfab_error("can't bind address", -ret);
		return -1;
	}

	iwprintf(5, "%s() calling rdma listen", __func__);


	#ifdef OF_SOFTWARE
	rdma_info->cma_listen_id->verbs->swinfo->rnic_hndl = -1;
	rdma_info->cma_listen_id = rdma_info->cma_id;
	rdma_info->rdma_evt_channel->next_event = RDMA_CM_EVENT_CONNECT_REQUEST;
	#endif

	ret = rdma_listen(rdma_info->cma_listen_id, 0);
	if (ret) {
		if (ret == -1)
			openfab_error("Can't listen", errno);
		else
			openfab_error("Can't listen", -ret);
		return -1;
	}

	#ifdef OF_SOFTWARE
	rdma_info->rdma_evt_channel->fd =
				rdma_info->cma_listen_id->verbs->swinfo->fd;
	#endif


	iwprintf(5, "%s() Exit", __func__);

    return 0;

}

int rdma_wait_connection(info *rdma_info)
/*Client function to wait until server tries to make rdma connection*/
{

	int ret;
	struct rdma_cm_event *event;
	struct rdma_conn_param conn_param = {
		.responder_resources = 1,
		.initiator_depth = 1,
		.retry_count = 5,
	};

	iwprintf(5, "%s() Entry", __func__);


	/*Magical Buf Fix  - Event channel must be associated with the wrong
	CMA ID from softwares point of view*/
	#ifdef OF_SOFTWARE
		rdma_info->rdma_evt_channel->cm_id = rdma_info->cma_id;
		rdma_info->cma_id->verbs->swinfo->fd = rdma_info->cma_listen_id->verbs->swinfo->fd;
	#endif

	iwprintf(5, "%s() Waiting for a CM event", __func__);

	ret = rdma_get_cm_event(rdma_info->rdma_evt_channel, &event);
	if (ret) {
		if (ret == -1){
			openfab_error("Unable to get a CM event", errno);
			return -1;
		} else{
			openfab_error("Unable to get a CM event", -ret);
			return -1;
		}
	}

	iwprintf(5, "%s() got CM event now to check and ack it", __func__);

	if(event->event != RDMA_CM_EVENT_CONNECT_REQUEST){
		iwprintf(0,"inavlid event received, %d, recevd %d",
				RDMA_CM_EVENT_CONNECT_REQUEST, event->event);
		openfab_error("Invalid event received", errno);
		return -1;
	}

	rdma_info->cma_id = event->id;  /*need to remember this for later its
					where the real connection is*/
	rdma_info->cma_id->qp = rdma_info->qp_hndl;

	ret = rdma_ack_cm_event(event);
	if(ret){
		openfab_error("Unable to release event", -ret);
		return -1;
	}

	iwprintf(5, "%s() going to accept conenction", __func__);

	ret = rdma_accept(rdma_info->cma_id, &conn_param);
	if (ret) {
		if (ret == -1)
			openfab_error("can not accept connection ret was -1",
						errno);
		else
			openfab_error("can't accept conn ret is not -1 though",
						-ret);
		return -1;
	}


	iwprintf(5, "%s() Connection accepted called, for wait for CM event",
				__func__);

	#ifdef OF_SOFTWARE
		rdma_info->rdma_evt_channel->next_event =
					RDMA_CM_EVENT_ESTABLISHED;
	#endif

	/*now check for connection made event*/
	ret = rdma_get_cm_event(rdma_info->rdma_evt_channel, &event);
	if (ret) {
		if (ret == -1){
			openfab_error("Unable to get a CM event", errno);
			return -1;
		}
		else{
			openfab_error("Unable to get a CM event", -ret);
			return -1;
		}
	}

	if(event->event != RDMA_CM_EVENT_ESTABLISHED){
		iwprintf(0,"inavlid event received, expecting  %d RDMA_CM_EVENT_ESTABLISHED, recevd %d", RDMA_CM_EVENT_ESTABLISHED, event->event);
		openfab_error("Invalid event received", errno);
		return -1;
	}

	iwprintf(5, "%s() Got CM event for accept", __func__);

	/*clear connection made event*/
	ret = rdma_ack_cm_event(event);
	if (ret){
		openfab_error("Unable to release event", -ret);
		return -1;
	}


	rdma_info->rnic_state = CONNECTED;

	iwprintf(5, "%s() Exit", __func__);

	return 0;

}








int rdma_send_signal(info *rdma_info)
/*Client function to send notice to server that RDMA read is finished*/
{
	int ret;
	struct ibv_send_wr *bad_wr;


	iwprintf(5, "%s() Entry", __func__);

	rdma_info->send_buf = malloc(SIGNAL_SIZE);
	if (!rdma_info->send_buf){
		openfab_error("Unable to malloc send buf", errno);
		return -1;
	}
	rdma_info->send_malloced = 1;


	/*register send buffer first*/
	rdma_info->send_mr = ibv_reg_mr(rdma_info->prot_id, rdma_info->send_buf,
			SIGNAL_SIZE,
			IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
	if (!rdma_info->send_mr){
		openfab_error("Unable to register rdma buffer", errno);
		return -1;
	}
	rdma_info->send_reg = 1;

	rdma_info->send_sge.addr = uint64_from_ptr(rdma_info->send_buf);
	rdma_info->send_sge.length = SIGNAL_SIZE;
	rdma_info->send_sge.lkey = rdma_info->send_mr->lkey;

	memset(&(rdma_info->send_wr), 0, sizeof(rdma_info->send_wr));
	rdma_info->send_wr.wr_id = uint64_from_ptr(&rdma_info->send_wr);
	rdma_info->send_wr.sg_list = &(rdma_info->send_sge);
	rdma_info->send_wr.num_sge = 1;
	rdma_info->send_wr.opcode = IBV_WR_SEND;
	rdma_info->send_wr.send_flags = IBV_SEND_SIGNALED;

	ret = ibv_post_send(rdma_info->qp_hndl, &(rdma_info->send_wr), &bad_wr);
	if (ret){
		openfab_error("Unable to post send", ret);
		return -1;
	}

	iwprintf(5, "%s() Exit", __func__);

	return 0;
}



int rdma_client_cleanup( info *rdma_info)
/*Client function to tear down connection*/
{
	int ret;
	int err_val = 0;
	struct rdma_cm_event *event;


	sleep(1);

	iwprintf(5, "%s() Disconnecting", __func__);
	ret = rdma_disconnect(rdma_info->cma_id);
	if (ret) {
		openfab_error("Unable to disconnect", errno);
		//~ err_val = -1;
		return 0;
	}

	/*Get disconnect event*/
	rdma_get_cm_event(rdma_info->rdma_evt_channel, &event);
	if (event->event != RDMA_CM_EVENT_DISCONNECTED) {
		openfab_error("Unable to get disconnected event", errno);
		err_val = -1;
	} else {
		iwprintf(5, "%s() Got disconnect event", __func__);
	}

	ret = rdma_ack_cm_event(event);
	if(ret) {
		openfab_error("Unable to release event", -ret);
		return -1;
	}

	rdma_destroy_qp(rdma_info->cma_id);

	/*XXXX TODO: Finish tearing down RDMA related stuff*/



	return err_val;
}





int rdma_allocate_client_resources(info *rdma_info, uint32_t wget_rdma_bufsize)
/*Client function to allocate a big giant static buffer to receive requests*/
{

	iwprintf(5, "%s() Entry", __func__);


	rdma_info->rdma_buf = malloc(wget_rdma_bufsize);
	if (!rdma_info->rdma_buf){
		openfab_error("Can't allocate memory", errno);
		return -1;
	}
	rdma_info->rdma_malloced = 1;

	/*register rdma buffer*/
	rdma_info->rdma_mr = ibv_reg_mr(rdma_info->prot_id, rdma_info->rdma_buf,
			wget_rdma_bufsize,
			IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
	if (!rdma_info->rdma_mr){
		openfab_error("Unable to register rdma buffer", errno);
		return -1;
	}
	rdma_info->rdma_reg = 1;

	rdma_info->rdma_stag_index = rdma_info->rdma_mr->rkey;
	rdma_info->rdma_to = uint64_from_ptr(rdma_info->rdma_buf);

	#ifdef __x86_64__
	iwprintf(5,"%s() Buffer Info STag = %d and TO = %ld",
				__func__,
				rdma_info->rdma_stag_index,
				rdma_info->rdma_to);
	#else
	iwprintf(5,"%s() Buffer Info STag = %d and TO = %lld",
				__func__,
				rdma_info->rdma_stag_index,
				rdma_info->rdma_to);
	#endif


	iwprintf(5, "%s() Exit", __func__);

	return 0;
}

int rdma_read_file(IGNORE info *rdma_info, IGNORE uint32_t size)
/*Client function to do an RDMA read from server into buffer*/
{

	/*build work request to do the RDMA read*/

	fprintf(stderr, "PLACEHOLDER FOR CODE TO BUILD RDMA READ REQUEST\n");

	/*actually do the RDMA read*/

	fprintf(stderr, "THEN WE DO THE RDMA READ\n");

	fprintf(stderr, "THEN NEED TO POLL OFF THE CQ AND THE FILE WILL BE IN MEMORY\n");


	return 0;
}
