/*
 Apache httpd module that uses RDMA to deliver data.

 This code is made available via the Apache License v2.0 (see LICENSE)
 and is Copyright 2007
		Dennis Dalessandro (dennis@osc.edu) - Ohio Supercomputer Center
		Pete Wyckoff (pw@osc.edu) - Ohio Supercomputer Center


 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <libgen.h>
#include <sys/errno.h>
#include <netdb.h>
#include <time.h>
#include "rdma/rdma_cma.h"
#include "infiniband/verbs.h"

#include "perfmon.h"
#include "pipeline_write.h"

#define uint64_from_ptr(p) ((uint64_t)(uintptr_t)(p))

#define CONTROL_BUF_SIZE 1024
#define CONN_BACK_LOG 10
#define MAX_POLL 10000000
#define SIGNAL_SIZE 10

#if 0
#define IW_VERBOSE_LEVEL 2
#define iwprintf(lvl,fmt,args...) \
    do { \
		if (lvl <= IW_VERBOSE_LEVEL) \
	    	iw_print_info(fmt,##args); \
    } while (0)
#else
#  define iwprintf(lvl,fmt,...) do { } while (0)
#endif

#define info OFINFO

//~ #define SENDFILE

typedef enum {
	IDLE = 1,
	CONNECTED,
	OPEN,
	CLOSED,
} state_t;

typedef struct {
	state_t				rnic_state;
	struct ibv_context 		*ibv_hndl;
	struct rdma_event_channel 	*rdma_evt_channel;
	struct rdma_cm_id 		*cma_listen_id;
	struct rdma_cm_id 		*cma_id;
	struct ibv_comp_channel  	*cq_channel;
	struct ibv_pd 			*prot_id;
	struct ibv_cq 			*cq_hndl;
	struct ibv_qp 			*qp_hndl;
	struct ibv_mr 			*rdma_mr;
	struct ibv_mr 			*recv_mr;
	struct ibv_mr 			*send_mr;
	struct ibv_sge 			rdma_sge;
	struct ibv_sge 			recv_sge;
	struct ibv_sge 			send_sge;
	struct ibv_send_wr  		rdma_wr;
	struct ibv_send_wr  		send_wr;
	struct ibv_recv_wr  		recv_wr;
	unsigned short			port;
	uint32_t			rdma_stag_index;
	uint32_t			remote_stag;
	uint64_t			remote_to;
	void				*rdma_buf;
	void				*recv_buf;
	void				*send_buf;
	uint8_t				rdma_mapped;
	uint8_t				rdma_malloced;
	uint8_t				recv_malloced;
	uint8_t				send_malloced;
	uint8_t				rdma_reg;
	uint8_t				recv_reg;
	uint8_t				send_reg;
	uint32_t			cq_count;
	uint32_t			rdma_bufsize;
	uint64_t			rdma_to;
	char 				ip_address[20];
	int 				not_tcp;  /*set this if its not TCP*/
	int 				port_override;  /*set to 0 if don't want to use, set before calling rdma init*/
	uint32_t			maxlen;  /*clients max len*/
	int				called_close;
	#ifdef SENDFILE
	rsf_desc_t 			rsf_descriptor;
	#endif
} info;

/*server side and shared functions*/
int rdma_zero_info(info *rdma_info);
int rdma_init(info *rdma_info);
void iw_print_info(const char *fmt, ...);
int rdma_do_connect(info *rdma_info, char *server, short unsigned int port);
int rdma_reap_completions(info *rdma_info, uint32_t num_events);
int rdma_deallocate_resources(info *rdma_info);
int rdma_cleanup(info *rdma_info);
const char *my_openib_wc_status_string(int status);
void openfab_error(const char *msg, int code);
int rdma_write_from_file(info *rdma_info, int fd, int offset, int len);
int setup_read_buffer_file(info *rdma_info, int fd, int offset, int len);
int rdma_write_from_buffer(info *rdma_info, void *buffer, int len);
int rdma_client_cleanup(info *rdma_info);

/*client side only functions*/
int rdma_allocate_client_resources(info *rdma_info, uint32_t wget_rdma_bufsize);
int rdma_listen_connection(info *rdma_info);
int rdma_wait_connection(info *rdma_info);

/*client RDMA read*/
int rdma_read_file(info *rdma_info, uint32_t size);
int rdma_send_signal(info *rdma_info);



